#include "RL_ToneSongPlayer.h"

const unsigned char buzzerPin = 8;
const unsigned int songTempo = 114;
const PROGMEM uint16_t song[] = {
  // Never Gonna Give You Up - Rick Astley
  // Score available at https://musescore.com/chlorondria_5/never-gonna-give-you-up_alto-sax
  // Arranged by Chlorondria

  NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,

  NOTE_E5,-8, NOTE_E5,-8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,-8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16, //18
  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,8, NOTE_A4,8, NOTE_A4,8, 
  NOTE_E5,4, NOTE_D5,2, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_FS5,-8, NOTE_FS5,-8, NOTE_E5,-4, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,
  NOTE_A5,4, NOTE_CS5,8, NOTE_D5,-8, NOTE_CS5,16, NOTE_B4,8, NOTE_A4,16, NOTE_B4,16, NOTE_D5,16, NOTE_B4,16,

  NOTE_D5,4, NOTE_E5,8, NOTE_CS5,-8, NOTE_B4,16, NOTE_A4,4, NOTE_A4,8,  //23
  NOTE_E5,4, NOTE_D5,2, REST,4,
};

// Create new song player
RL_ToneSongPlayer songPlayer(buzzerPin, song, sizeof(song), songTempo, 1);

// Song player event handlers

void onSongTone ( RL_ToneSongPlayer * player, unsigned char pin )
{
  Serial.println("onSongTone(): Frequency " + String(player->toneFrequency) + " for note " + String(player->noteNum) + " of " + String(player->numNotes));
}

void onSongLoop ( unsigned int timesLooped )
{
  Serial.println("onSongLoop(): Looped " + String(timesLooped) + " time(s)");
}

void onSongEnd ()
{
  Serial.println("onSongEnd(): Music over");
}

void onSongProgress ( RL_ToneSongPlayer * player )
{
  // If playing a tone, do a corresponding action
  if ( player->toneFrequency > 0 )
  {
    // If the range should end before the tone, put that offset here
    // For example, if you wanted to finish an action 50 milliseconds before the tone ends, you would enter 50
    unsigned int toneDurationOffsetMs = 0;
    // How long the action has with the current tone
    unsigned int actionDuration = player->currentToneDurationMs - toneDurationOffsetMs;
    // The action progress inverse of the tone progress
    float actionDurationProgress = 1.0 - player->currentToneProgress;

    // For example:
    // Update LED brightness to fade during tone duration
    //unsigned char ledBrightness = actionDurationProgress * 255;
    //digitalWrite(ledPin, ledBrightness);//digitalWrite:HIGH,LOW or use a PWM pin with a range from 0-255
    //Serial.println("LED: " + String(actionDurationProgress) + " " + String(ledBrightness));
  }

  // Throttle Serial for SRAM memory optimization
  unsigned long nowMs = millis();
  static unsigned long lastRunTimeMs = 0;
  unsigned int throttleForMs = 250;
  if ( nowMs - lastRunTimeMs >= throttleForMs )
  {
    lastRunTimeMs = nowMs;

    Serial.println("onSongProgress(): Tone progress " + String(player->currentToneProgress) + 
    " or " + String(player->currentTonePlayedMs) + "/" + String(player->currentToneDurationMs) + "ms" + 
    " for note " + String(player->noteNum) + "/" + String(player->numNotes) +
    ", Song progress " + String(player->songProgress) + 
    " or " + String(player->songPlayedMs) + "/" + String(player->songDurationMs) + "ms");
  }
}

void setup() {
  // Begin a serial connection for debugging
  Serial.begin(9600);

  // Initialize music player
  songPlayer.play();
  // Attach event handlers
  songPlayer.setOnProgress(onSongProgress);
  songPlayer.setOnTone(onSongTone);
  songPlayer.setOnLoop(onSongLoop);
  songPlayer.setOnEnd(onSongEnd);
}

void loop() {
  // Keep song player running
  songPlayer.update();
}
