/*!
*  @file RL_ToneSongPlayer.h
*
*  @mainpage Arduino library to play songs composed of tone notes and durations on a digital pin in a non-interrupting way without using the delay() function.
*
*  @section intro_sec Introduction
*
*  This is an Arduino library to play songs composed of tone notes and durations on a digital pin in a non-interrupting way without using the delay() function. Features include event callbacks, looping, and progress tracking.
*
*  @section author Joshua Logsdon
*
*  Written by Joshua Logsdon
*
*  @section license License
*
*  MIT license, all text above must be included in any redistribution
*/

// The include guard prevents program errors if the library is accidentally included more than once in the sketch
// Begin include guard
#ifndef RL_ToneSongPlayer_H
#define RL_ToneSongPlayer_H

// The core library that contains all of the Arduino’s built-in functions
#include "Arduino.h"

/*==============================================================================
* Public constants
*=============================================================================*/

// Tone notes, constants so they can be used for songs outside of the library
#define REST 0
#define NOTE_REST REST
#define NOTE_B0 31
#define NOTE_C1 33
#define NOTE_CS1 35
#define NOTE_D1 37
#define NOTE_DS1 39
#define NOTE_E1 41
#define NOTE_F1 44
#define NOTE_FS1 46
#define NOTE_G1 49
#define NOTE_GS1 52
#define NOTE_A1 55
#define NOTE_AS1 58
#define NOTE_B1 62
#define NOTE_C2 65
#define NOTE_CS2 69
#define NOTE_D2 73
#define NOTE_DS2 78
#define NOTE_E2 82
#define NOTE_F2 87
#define NOTE_FS2 93
#define NOTE_G2 98
#define NOTE_GS2 104
#define NOTE_A2 110
#define NOTE_AS2 117
#define NOTE_B2 123
#define NOTE_C3 131
#define NOTE_CS3 139
#define NOTE_D3 147
#define NOTE_DS3 156
#define NOTE_E3 165
#define NOTE_F3 175
#define NOTE_FS3 185
#define NOTE_G3 196
#define NOTE_GS3 208
#define NOTE_A3 220
#define NOTE_AS3 233
#define NOTE_B3 247
#define NOTE_C4 262
#define NOTE_CS4 277
#define NOTE_D4 294
#define NOTE_DS4 311
#define NOTE_E4 330
#define NOTE_F4 349
#define NOTE_FS4 370
#define NOTE_G4 392
#define NOTE_GS4 415
#define NOTE_A4 440
#define NOTE_AS4 466
#define NOTE_B4 494
#define NOTE_C5 523
#define NOTE_CS5 554
#define NOTE_D5 587
#define NOTE_DS5 622
#define NOTE_E5 659
#define NOTE_F5 698
#define NOTE_FS5 740
#define NOTE_G5 784
#define NOTE_GS5 831
#define NOTE_A5 880
#define NOTE_AS5 932
#define NOTE_B5 988
#define NOTE_C6 1047
#define NOTE_CS6 1109
#define NOTE_D6 1175
#define NOTE_DS6 1245
#define NOTE_E6 1319
#define NOTE_F6 1397
#define NOTE_FS6 1480
#define NOTE_G6 1568
#define NOTE_GS6 1661
#define NOTE_A6 1760
#define NOTE_AS6 1865
#define NOTE_B6 1976
#define NOTE_C7 2093
#define NOTE_CS7 2217
#define NOTE_D7 2349
#define NOTE_DS7 2489
#define NOTE_E7 2637
#define NOTE_F7 2794
#define NOTE_FS7 2960
#define NOTE_G7 3136
#define NOTE_GS7 3322
#define NOTE_A7 3520
#define NOTE_AS7 3729
#define NOTE_B7 3951
#define NOTE_C8 4186
#define NOTE_CS8 4435
#define NOTE_D8 4699
#define NOTE_DS8 4978

// Finish out same note definitions
// https://gist.github.com/mikeputnam/2820675?permalink_comment_id=3263645#gistcomment-3263645
// Sharps
#define NOTE_BS0 NOTE_C1
#define NOTE_BS1 NOTE_C2
#define NOTE_BS2 NOTE_C3
#define NOTE_BS3 NOTE_C4
#define NOTE_BS4 NOTE_C5
#define NOTE_BS5 NOTE_C6
#define NOTE_BS6 NOTE_C7
#define NOTE_BS7 NOTE_C8
#define NOTE_ES1 NOTE_F1
#define NOTE_ES2 NOTE_F2
#define NOTE_ES3 NOTE_F3
#define NOTE_ES4 NOTE_F4
#define NOTE_ES5 NOTE_F5
#define NOTE_ES6 NOTE_F6
#define NOTE_ES7 NOTE_F7
// Flats
#define NOTE_AF1 NOTE_BS1
#define NOTE_AF2 NOTE_BS2
#define NOTE_AF3 NOTE_BS3
#define NOTE_AF4 NOTE_BS4
#define NOTE_AF5 NOTE_BS5
#define NOTE_AF6 NOTE_BS6
#define NOTE_AF7 NOTE_BS7
#define NOTE_BF1 NOTE_AS1
#define NOTE_BF2 NOTE_AS2
#define NOTE_BF3 NOTE_AS3
#define NOTE_BF4 NOTE_AS4
#define NOTE_BF5 NOTE_AS5
#define NOTE_BF6 NOTE_AS6
#define NOTE_BF7 NOTE_AS7
#define NOTE_CF1 NOTE_B0
#define NOTE_CF2 NOTE_B1
#define NOTE_CF3 NOTE_B2
#define NOTE_CF4 NOTE_B3
#define NOTE_CF5 NOTE_B4
#define NOTE_CF6 NOTE_B5
#define NOTE_CF7 NOTE_B6
#define NOTE_CF8 NOTE_B7
#define NOTE_DF1 NOTE_CS1
#define NOTE_DF2 NOTE_CS2
#define NOTE_DF3 NOTE_CS3
#define NOTE_DF4 NOTE_CS4
#define NOTE_DF5 NOTE_CS5
#define NOTE_DF6 NOTE_CS6
#define NOTE_DF7 NOTE_CS7
#define NOTE_DF8 NOTE_CS8
#define NOTE_EF1 NOTE_DS1
#define NOTE_EF2 NOTE_DS2
#define NOTE_EF3 NOTE_DS3
#define NOTE_EF4 NOTE_DS4
#define NOTE_EF5 NOTE_DS5
#define NOTE_EF6 NOTE_DS6
#define NOTE_EF7 NOTE_DS7
#define NOTE_FF1 NOTE_E1
#define NOTE_FF2 NOTE_E2
#define NOTE_FF3 NOTE_E3
#define NOTE_FF4 NOTE_E4
#define NOTE_FF5 NOTE_E5
#define NOTE_FF6 NOTE_E6
#define NOTE_FF7 NOTE_E7
#define NOTE_GF1 NOTE_FS1
#define NOTE_GF2 NOTE_FS2
#define NOTE_GF3 NOTE_FS3
#define NOTE_GF4 NOTE_FS4
#define NOTE_GF5 NOTE_FS5
#define NOTE_GF6 NOTE_FS6
#define NOTE_GF7 NOTE_FS7

/*!
*  @brief  Class definition
*/
class RL_ToneSongPlayer
{
  /*==============================================================================
  * Private properties
  *=============================================================================*/
  
  private:
    unsigned long _currentToneStartMs;
    unsigned char _pin;
    unsigned long _playStartMs;
    uint16_t *_song;
    int _songNoteIndex = -1;

  /*==============================================================================
  * Public properties
  *=============================================================================*/

  public:
    unsigned int currentToneDurationMs;
    int currentTonePlayedMs;
    float currentToneProgress;
    unsigned char debug = 0;
    bool isMuted;
    bool isPaused;
    bool isPlaying;
    unsigned char loop;
    unsigned int noteNum = 0;
    unsigned int numNotes = 0;
    unsigned long songDurationMs = 0;
    unsigned long songPlayedMs = 0;
    float songProgress;
    unsigned int songSize;
    unsigned long songTonesPlayedCompleteMs = 0;
    unsigned int tempo;
    unsigned int timesLooped = 0;
    int toneFrequency = -1;
    // Callback variable function definitions
    void (*onEndCallback)(void);
    void (*onLoopCallback)(unsigned int timesLooped);
    void (*onProgressCallback)(RL_ToneSongPlayer * player);
    void (*onToneCallback)(RL_ToneSongPlayer * player, unsigned char pin);

    /*==============================================================================
    * Constructor
    *=============================================================================*/

    /*!
    *  @brief  Instantiates a new class
    *  @param  pin
    *          Pin number connected to piezo buzzer
    *  @param  song
    *          Song array set in the sketch like: const PROGMEM uint16_t songTest[] = {...Notes and durations here...};
    *  @param  songSize
    *          sizeof(song) passed at class instantiation from the sketch
    *          Because song is passed in as a pointer, the size has to be passed from outside of the class
    *  @param  tempo
    *          The song tempo
    *  @param  loop
    *          How many times to loop the song, -1 being infinite
    *  @return void
    */
    RL_ToneSongPlayer ( int pin, const uint16_t* song, int songSize, int tempo, int loop = 0 )
    {
      setPin(pin);
      this->tempo = tempo;
      this->loop = loop;
      setSongSize(songSize);
      setSong(song);
    }

    /*==============================================================================
    * Getter and setter methods
    *=============================================================================*/

    void setPin ( unsigned char pin )
    {
      _pin = pin;
    }

    unsigned char getPin ()
    {
      return _pin;
    }

    void setSong ( const uint16_t* song )
    {
      _song = song;

      // Calculate song duration
      songDurationMs = 0;
      unsigned int songArrLen = numNotes * 2;
      for ( unsigned int i = 0; i <= songArrLen; i += 2 )
      {
        songDurationMs += getSongNoteDuration(i, tempo);
      }
    }

    uint16_t getSong ()
    {
      return _song;
    }

    void setSongSize ( int songSize )
    {
      this->songSize = songSize;
      numNotes = getNumNotes();
    }

    int getSongSize ()
    {
      return songSize;
    }
    
    void setSongNoteIndex ( unsigned long songNoteIndex = 0 )
    {
      _songNoteIndex = songNoteIndex;
      // Song arrays have note and duration values, so the array index is * 2 of the actual note number being played
      // Array indexes start at 0 so offset it by adding 1
      noteNum = (songNoteIndex / 2) + 1;
    }

    unsigned long getSongNoteIndex ()
    {
      return _songNoteIndex;
    }
    
    void setNoteIndex ( unsigned long noteIndex = 0 )
    {
      setSongNoteIndex(noteIndex * 2);
    }
    
    void setOnEnd( void (*funct)(void) )
    {
      onEndCallback = funct;
    }
    
    void setOnLoop( void (*funct)(void) )
    {
      onLoopCallback = funct;
    }
    
    void setOnProgress( void (*funct)(void) )
    {
      onProgressCallback = funct;
    }
    
    void setOnTone( void (*funct)(void) )
    {
      onToneCallback = funct;
    }

    /*==============================================================================
    * Methods
    *=============================================================================*/

    void mute ( bool mute = true )
    {
      isMuted = mute;
      
      // If muted, stop tone
      if ( isMuted )
      {
        noTone(_pin);
      }
    }

    void pause ()
    {
      isPaused = true;
      isPlaying = false;
      // Stop tone
      noTone(_pin);
    }

    void play ()
    {
      isPaused = false;
      isPlaying = true;
      _playStartMs = millis();
    }

    void seek ( unsigned long noteIndex = 0 )
    {
      setNoteIndex(noteIndex);
      
      // Reset to no tone being played to force queueing up song note index
      toneFrequency = -1;
      currentToneDurationMs = 0;

      // Reset time song played
      songPlayedMs = 0;
      // If starting past the beginning
      if ( noteIndex > 0 )
      {
        // Calculate song played duration by iterating through tone durations backwards to the song beginning
        for ( unsigned int i = noteIndex * 2; i >= 0; i -= 2 )
        {
          songPlayedMs += getSongNoteDuration(i, tempo);
        }
      }

      songTonesPlayedCompleteMs = songPlayedMs;
    }

    void stop ()
    {
      isPaused = false;
      isPlaying = false;
      // Stop tone
      noTone(_pin);
      // Reset for play
      seek(0);
    }

    void toggleMute ()
    {
      mute(! isMuted);
    }

    int getNoteTypeDuration ( int noteType, int tempo )
    {
      // this calculates the duration of a whole note in ms
      int wholenote = (60000 * 4) / tempo;

      // notes of the moledy followed by the duration.
      // 1 = whole note | 2 = half note | 4 = quarter note | 
      // 8 = eighteenth | 16 = sixteenth | 32 = thirty-second, etc.
      // a 4 means a quarter note, 8 an eighteenth , 16 sixteenth, so on
      // !!negative numbers are used to represent dotted notes,
      // so -4 means a dotted quarter note, that is, a quarter plus an eighteenth!!
      
      // calculates the duration of each note
      //int noteType = pgm_read_word(&song[songNoteIndex + 1]);
      int noteDuration = 0;
      
      if ( noteType > 0 ) {
        // regular note, just proceed
        noteDuration = (wholenote) / noteType;
      }
      else if ( noteType < 0 )
      {
        // dotted notes are represented with negative durations!!
        noteDuration = (wholenote) / abs(noteType);
        noteDuration *= 1.5; // increases the duration in half for dotted notes
      }

      return noteDuration;
    }

    int getNumNotes ()
    {
      // Because song is passed in as a pointer, the size has to be passed from outside of the class
      // sizeof(_song[0]) == 2
      return songSize / sizeof(_song[0]) / 2;

      // sizeof gives the number of bytes, each int value is composed of two bytes (16 bits)
      // there are two values per note (pitch and duration), so for each note there are four bytes
      //return sizeof(song) / sizeof(song[0]) / 2;
    }
    
    int getSongNoteDuration ( int songNoteIndex, int tempo )
    {
      int noteType = pgm_read_word(&_song[songNoteIndex + 1]);
      int noteDuration = getNoteTypeDuration(noteType, tempo);
      if ( debug ) Serial.println( "RL_ToneSongPlayer: getSongNoteDuration(): " + String(songNoteIndex) + " " + String(tempo) + " " + String(noteType) + " " + String(noteDuration));
      return noteDuration;
    }

    /*!
    *  @brief  Code to run in sketch loop to keep player running
    */
    void update ()
    {
      if ( debug ) Serial.println("RL_ToneSongPlayer: update() " + String(isPlaying));

      // If not playing, stop
      if ( ! isPlaying )
      {
        return;
      }

      unsigned long nowMs = millis();

      // Keep value within range, even if no tone has been played yet (-1)
      currentTonePlayedMs = toneFrequency > -1 ? nowMs - _currentToneStartMs : 0;
      currentTonePlayedMs = constrain(currentTonePlayedMs, 0, currentToneDurationMs);
      
      // Update song played from song tones played complete ms and current tone played ms
      songPlayedMs = songTonesPlayedCompleteMs + currentTonePlayedMs;
      
      // Update song progress from played ms and song duration ms
      songProgress = (float)songPlayedMs / (float)(songDurationMs > 0 ? songDurationMs : 1);

      // Keep progress in valid range
      currentToneProgress = 0;
      // If denominator not 0 and nominator less than denominator, get progress percent from 0.0 to 1.0
      if ( currentToneDurationMs > 0 )
      {
        currentToneProgress = (float)currentTonePlayedMs / (float)currentToneDurationMs;
      }
      
      // If a note is playing and progress callback
      if ( _songNoteIndex > -1 && onProgressCallback )
      {
        onProgressCallback(this);
      }

      // If tone has not played for note length (or possibly at song start with inialized values), stop
      if ( currentTonePlayedMs < currentToneDurationMs )
      {
        return;
      }

      // Update song tones played complete ms with tone duration ms
      songTonesPlayedCompleteMs += currentToneDurationMs;

      if ( debug ) Serial.println( "RL_ToneSongPlayer: noteNum " + String(noteNum) + " numNotes " + String(numNotes) );
      
      // If past that last note, check loop or stop
      if ( noteNum >= numNotes )
      {
        // If looping infinitely or a certain number of times
        if ( loop < 0 || timesLooped < loop )
        {
          // Reset song index, increment loop times
          seek(0);
          timesLooped++;
          
          if ( onLoopCallback )
          {
            onLoopCallback(timesLooped);
          }
        }
        // Else stop
        else
        {
          stop();

          if ( onEndCallback )
          {
            onEndCallback();
          }

          return;
        }
      }

      // Play next tone
      
      // Move on to next note in song array, or start at beginning if not set
      setSongNoteIndex(_songNoteIndex < 0 ? 0 : _songNoteIndex + 2);

      // Update when tone stared playing, then play tone
      _currentToneStartMs = nowMs;

      // Determine how long to play tone
      currentToneDurationMs = getSongNoteDuration(_songNoteIndex, tempo);
      toneFrequency = pgm_read_word(&_song[_songNoteIndex]);
      
      if ( debug ) Serial.println( "RL_ToneSongPlayer: Tone: " + String(_songNoteIndex) + " " + String(toneFrequency) + " " + String(currentToneDurationMs) );
      
      // Play tone if not muted
      if ( ! isMuted )
      {
        // Stop tone before playing new tone
        noTone(_pin);
        tone(_pin, toneFrequency, currentToneDurationMs);
      }

      if ( onToneCallback )
      {
        onToneCallback(this, _pin);
      }
    }
};

// End include guard
#endif
